package eu.mhsl.craftattack.spawn.appliances.settings;

import eu.mhsl.craftattack.spawn.Main;
import eu.mhsl.craftattack.spawn.appliance.Appliance;
import eu.mhsl.craftattack.spawn.appliance.ApplianceCommand;
import eu.mhsl.craftattack.spawn.appliances.settings.datatypes.Setting;
import eu.mhsl.craftattack.spawn.appliances.settings.listeners.OpenSettingsShortcutListener;
import eu.mhsl.craftattack.spawn.appliances.settings.listeners.SettingsInventoryListener;
import eu.mhsl.craftattack.spawn.appliances.settings.settings.*;
import net.kyori.adventure.text.Component;
import org.bukkit.Bukkit;
import org.bukkit.entity.Player;
import org.bukkit.event.Listener;
import org.bukkit.inventory.Inventory;
import org.jetbrains.annotations.NotNull;

import java.util.Arrays;
import java.util.List;
import java.util.WeakHashMap;
import java.util.concurrent.atomic.AtomicInteger;

public class Settings extends Appliance {
    private static Settings settingsInstance;

    public enum Key {
        TechnicalTab,
        ShowJoinAndLeaveMessages,
        EnablePortableCrafting,
        EnableSettingsShortcut,
        AutoShulker,
        SignEdit,
        HotbarReplacer,
        ChatMentions,
        DoubleDoors,
    }

    public static Settings instance() {
        if(settingsInstance != null) return settingsInstance;
        Settings instance = Main.instance().getAppliance(Settings.class);
        Settings.settingsInstance = instance;
        return instance;
    }

    public record OpenSettingsInventory(Inventory inventory, List<Setting<?>> settings) {
    }

    private final WeakHashMap<Player, OpenSettingsInventory> openSettingsInventories = new WeakHashMap<>();
    private final WeakHashMap<Player, List<Setting<?>>> settingsCache = new WeakHashMap<>();

    private List<Setting<?>> getSettings(Player player) {
        if(settingsCache.containsKey(player)) return settingsCache.get(player);

        List<Setting<?>> settings = List.of(
            new PortableCraftingSetting(),
            new AutoShulkerSetting(),
            new SignEditSetting(),
            new HotbarReplaceSetting(),
            new ChatMentionSetting(),
            new ShowJoinAndLeaveMessagesSetting(),
            new TechnicalTablistSetting(),
            new SettingsShortcutSetting(),
            new DoubleDoorSetting()
        );

        settings.forEach(setting -> setting.initializeFromPlayer(player));
        this.settingsCache.put(player, settings);
        return settings;
    }

    public <T> T getSetting(Player player, Key key, Class<T> clazz) {
        Setting<?> setting = getSettings(player).stream()
            .filter(s -> s.getKey().equals(key))
            .findFirst()
            .orElseThrow();

        if(!clazz.equals(setting.dataType()))
            throw new IllegalStateException("Tried to retrieve Setting with Datatype " + clazz.getSimpleName() + " but expected " + setting.dataType().getSimpleName());
        if(!clazz.isInstance(setting.state()))
            throw new ClassCastException(clazz.getSimpleName() + " is not an instance of " + setting.dataType().getSimpleName());
        return clazz.cast(setting.state());
    }

    public void openSettings(Player player) {
        List<Setting<?>> settings = getSettings(player);
        Inventory inventory = Bukkit.createInventory(null, calculateInvSize(settings), Component.text("Einstellungen"));

        AtomicInteger row = new AtomicInteger(0);
        Arrays.stream(SettingCategory.values())
            .forEach(category -> {
                List<Setting<?>> categorizedSettings = settings.stream()
                    .filter(setting -> setting instanceof CategorizedSetting)
                    .filter(setting -> ((CategorizedSetting) setting).category().equals(category))
                    .toList();

                for (int i = 0; i < categorizedSettings.size(); i++) {
                    int slot = row.get() * 9 + i % 9;
                    inventory.setItem(slot, categorizedSettings.get(i).buildItem());

                    if (i % 9 == 8) {
                        row.incrementAndGet();
                    }
                }
                row.incrementAndGet();
            });

        List<Setting<?>> uncategorizedSettings = settings.stream()
            .filter(setting -> !(setting instanceof CategorizedSetting))
            .toList();

        for (int i = 0; i < uncategorizedSettings.size(); i++) {
            int slot = row.get() * 9 + i % 9;
            inventory.setItem(slot, uncategorizedSettings.get(i).buildItem());

            if (i % 9 == 8) {
                row.incrementAndGet();
            }
        }

        player.openInventory(inventory);
        this.openSettingsInventories.put(player, new OpenSettingsInventory(inventory, settings));
    }

    private int calculateInvSize(List<Setting<?>> settings) {
        int countOfUncategorized = (int) settings.stream().filter(setting -> !(setting instanceof CategorizedSetting)).count();
        return Arrays.stream(SettingCategory.values())
            .map(settingCategory -> settings.stream()
                .filter(setting -> setting instanceof CategorizedSetting)
                .filter(setting -> ((CategorizedSetting) setting).category().equals(settingCategory))
                .count())
            .map(itemCount -> itemCount + countOfUncategorized)
            .map(itemCount -> (int) Math.ceil((double) itemCount / 9))
            .reduce(Integer::sum)
            .orElse(1) * 9;
    }

    public void onSettingsClose(Player player) {
        if(!openSettingsInventories.containsKey(player)) return;
        openSettingsInventories.remove(player);
        player.updateInventory();
    }

    public boolean hasSettingsNotOpen(Player player) {
        return !this.openSettingsInventories.containsKey(player);
    }

    public OpenSettingsInventory getOpenInventory(Player player) {
        if(hasSettingsNotOpen(player)) throw new RuntimeException("Cannot retrieve data from closed Settings inventory!");
        return this.openSettingsInventories.get(player);
    }

    @Override
    @NotNull
    protected List<Listener> eventHandlers() {
        return List.of(
            new SettingsInventoryListener(),
            new OpenSettingsShortcutListener()
        );
    }

    @Override
    @NotNull
    protected List<ApplianceCommand<?>> commands() {
        return List.of(new SettingsCommand());
    }
}
